import UIKit

class AppTabBarController: UITabBarController, UITabBarControllerDelegate, WalkthroughViewControllerDelegate {
    
    struct Constants {
        static let reuseIdentifier = String(describing: AppTabBarController.self)
        static let storyboardName = "Documents"
    }
    
    static func buildViewController() -> AppTabBarController {
        let controller = UIStoryboard(name: Constants.storyboardName,
                                      bundle: .main).instantiateViewController(withIdentifier: Constants.reuseIdentifier)
        return controller as! AppTabBarController
    }
    
    required init?(coder: NSCoder) {
        super.init(coder: coder)
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        self.delegate = self
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        if UserDefaults.standard.showWalktroughAtLaunch {
            perform(#selector(showWalkTrough), with: self, afterDelay: 0.88)
        }
    }
    
    let walkthroughs = [
        WalkthroughModel(title: "Quick Overview",
                         subtitle: "Quickly visualize important business metrics. The overview in the home tab shows the most important metrics to monitor how your business is doing in real time.",
                         icon: "analytics-icon"),
        WalkthroughModel(title: "Analytics",
                         subtitle: "Dive deep into charts to extract valuable insights and come up with data driven product initiatives, to boost the success of your business.",
                         icon: "bars-icon"),
        WalkthroughModel(title: "Dashboard Feeds",
                         subtitle: "View your sales feed, orders, customers, products and employees.",
                         icon: "activity-feed-icon"),
        WalkthroughModel(title: "Get Notified",
                         subtitle: "Receive notifications when critical situations occur to stay on top of everything important.",
                         icon: "bell-icon"),
    ]
    
    @objc private func showWalkTrough() {
        let walkthroughVC = self.walkthroughVC()
        walkthroughVC.delegate = self
        addChildViewControllerWithView(walkthroughVC)
    }
    
    fileprivate func walkthroughVC() -> WalkthroughViewController {
        let viewControllers = walkthroughs.map { ClassicWalkthroughViewController(model: $0, nibName: "ClassicWalkthroughViewController", bundle: nil) }
        return WalkthroughViewController(nibName: "WalkthroughViewController",
                                         bundle: nil,
                                         viewControllers: viewControllers)
    }
    
    func walkthroughViewControllerDidFinishFlow(_ vc: WalkthroughViewController) {
        UIView.transition(with: self.view, duration: 1, options: .transitionCrossDissolve, animations: {
            vc.view.removeFromSuperview()
        }, completion: nil)
    }
}

class AppTabBar: UITabBar {
    
    private var middleButton = UIButton()
    private var shapeLayer: CALayer?
    
    private func addShape() {
        let shapeLayer = CAShapeLayer()
        shapeLayer.path = createPath()
        shapeLayer.strokeColor = UIColor.lightGray.cgColor
        let tabBackColor = UIColor.barTintColor
        shapeLayer.fillColor = tabBackColor.cgColor
        shapeLayer.lineWidth = 0.5
        
        //The below 4 lines are for shadow above the bar. you can skip them if you do not want a shadow
        shapeLayer.shadowOffset = CGSize(width:0, height:0)
        shapeLayer.shadowRadius = 10
        shapeLayer.shadowColor = UIColor.gray.cgColor
        shapeLayer.shadowOpacity = 0.3
        
        if let oldShapeLayer = self.shapeLayer {
            self.layer.replaceSublayer(oldShapeLayer, with: shapeLayer)
        } else {
            self.layer.insertSublayer(shapeLayer, at: 0)
        }
        self.shapeLayer = shapeLayer
    }
    
    override func awakeFromNib() {
        super.awakeFromNib()
        setupMiddleButton()
    }
    
    override func hitTest(_ point: CGPoint, with event: UIEvent?) -> UIView? {
        if self.isHidden {
            return super.hitTest(point, with: event)
        }
        
        let from = point
        let to = middleButton.center
        
        return sqrt((from.x - to.x) * (from.x - to.x) + (from.y - to.y) * (from.y - to.y)) <= 120 ? middleButton : super.hitTest(point, with: event)
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
        middleButton.center = CGPoint(x: UIScreen.main.bounds.width / 2, y: 10)
    }
    
    @objc private func setupMiddleButton() {
      let image = UIImage(named: "add_doc_icon")
      middleButton.setImage(image, for: UIControl.State.normal)
      middleButton.contentVerticalAlignment = .fill
      middleButton.contentHorizontalAlignment = .fill
      middleButton.imageEdgeInsets = .zero
      middleButton.frame.size = CGSize(width: 100, height: 100)
      middleButton.layer.masksToBounds = true
      middleButton.layer.cornerRadius = 50

      middleButton.center = CGPoint(x: UIScreen.main.bounds.width / 2, y: 0)
      middleButton.addTarget(self, action: #selector(buttonPressed), for: .touchUpInside)
      addSubview(middleButton)
    }
    
    @objc private func buttonPressed() {
        if let tabBarController = self.window!.rootViewController as? UITabBarController {
            NotificationCenter.default.post(name: NSNotification.Name.MiddleButtonTapped, object: middleButton)
            
            if tabBarController.selectedIndex != 0 {
                UIView.animate(withDuration: 0.33) {
                    tabBarController.selectedIndex = 0
                }
            }
        }
    }
    
    override func draw(_ rect: CGRect) {
        self.addShape()
    }
    
    private func createPath() -> CGPath {
        let height: CGFloat = 120
        let path = UIBezierPath()
        let centerWidth = self.frame.width / 2

        path.move(to: CGPoint(x: 0, y: 0)) // start top left
        path.addLine(to: CGPoint(x: (centerWidth - height * 0.5), y: 0)) // the beginning of the trough

        path.addArc(withCenter: CGPoint(x: centerWidth, y: 0),
                    radius: height * 0.5,
                    startAngle: CGFloat.pi,
                    endAngle: 0,
                    clockwise: false)

        path.addLine(to: CGPoint(x: (centerWidth + height * 0.5), y: 0))
        path.addLine(to: CGPoint(x: self.frame.width, y: 0))
        path.addLine(to: CGPoint(x: self.frame.width, y: self.frame.height))
        path.addLine(to: CGPoint(x: 0, y: self.frame.height))
        path.close()

        return path.cgPath
    }
}
