import UIKit
import PDFKit

class PDFManager {
  
  class func pdfThumbnail(url: URL, width: CGFloat = 240) -> UIImage? {
    guard let data = try? Data(contentsOf: url),
          let page = PDFDocument(data: data)?.page(at: 0) else {
      return nil
    }
    
    let pageSize = page.bounds(for: .mediaBox)
    let pdfScale = width / pageSize.width
    
    let scale = UIScreen.main.scale * pdfScale
    let screenSize = CGSize(width: pageSize.width * scale,
                            height: pageSize.height * scale)
    
    return page.thumbnail(of: screenSize, for: .mediaBox)
  }
  
  class func generatePdfThumbnail(of thumbnailSize: CGSize, for documentUrl: URL) -> UIImage? {
    guard let data = try? Data(contentsOf: documentUrl),
          let page = PDFDocument(data: data)?.page(at: 0) else {
      return nil
    }

    return page.thumbnail(of: thumbnailSize, for: .trimBox)
  }
  
  class func getPDFPageToImage(page: PDFPage, pageIndex: Int) -> UIImage? {
    let pageSize = page.bounds(for: .mediaBox)
    let screenSize = CGSize(width: pageSize.width,
                            height: pageSize.height)

    return page.thumbnail(of: screenSize, for: .mediaBox)
  }
  
    class func createPDFDocument(from image: UIImage, localFileManager: LocalFileManager?, folder: AppConfigurator.Folder) {
    guard let scannerURL = localFileManager?.getFolderUrl(folder: folder) else { return }
    let urlPDFtoSave = scannerURL.appendingPathComponent(String.getDocumentName())
    let document = PDFDocument()
    
    if let pdfPage = PDFPage(image: image) {
      document.insert(pdfPage, at: 0)
    }
    do {
      try document.dataRepresentation()?.write(to: urlPDFtoSave)
    } catch {
      print(error.localizedDescription)
    }
  }
  
  class func createPDFDocument(from image: UIImage, urlPDFtoSave: URL) {
    let document = PDFDocument()
    
    if let pdfPage = PDFPage(image: image) {
      document.insert(pdfPage, at: 0)
    }
    do {
      try document.dataRepresentation()?.write(to: urlPDFtoSave)
    } catch {
      print(error.localizedDescription)
    }
  }
    
    class func createMultiPDFPage(from imageURLs: [URL], localFileManager: LocalFileManager?, folder: AppConfigurator.Folder, _ completion: @escaping () -> ()) {
        guard let scannerURL = localFileManager?.getFolderUrl(folder: folder) else { return }
        let urlPDFtoSave = scannerURL.appendingPathComponent(String.getDocumentName())
        
        if imageURLs.isEmpty { return }
        let document = PDFDocument()

        let dispatchGroup = DispatchGroup()
        
        // Iterate through image URLs
        for (index, imageURL) in imageURLs.enumerated() {
            dispatchGroup.enter()
            DispatchQueue.global().async {
                if let data = try? Data(contentsOf: imageURL),
                   let image = UIImage(data: data),
                   let pdfPage = PDFPage(image: image) {
                    DispatchQueue.main.async {
                        document.insert(pdfPage, at: index)
                        if document.pageCount == imageURLs.count {
                            do {
                                try document.dataRepresentation()?.write(to: urlPDFtoSave)
                                completion()
                            } catch {
                                print(error.localizedDescription) // handle error
                            }
                        }
                        dispatchGroup.leave()
                    }
                } else {
                    dispatchGroup.leave()
                }
            }
        }
        dispatchGroup.notify(queue: .main) {
            print("All images have been processed")
        }
    }
  
  class func appendPage(image: UIImage, pdfView: PDFView, documentURL: URL?) {
    guard let page = PDFPage(image: image),
          let documentURL = documentURL,
          let lastIndex = pdfView.document?.pageCount else  { return }
    pdfView.document?.insert(page, at: lastIndex)
    pdfView.document?.write(to: documentURL)
    pdfView.go(to: page)
    
    LocalFileManager().updateThumbnail(forURL: documentURL)
  }
    
class func appendPage(imageURL: URL, pdfView: PDFView, documentURL: URL?) {
        DispatchQueue.global().async {
            if let data = try? Data(contentsOf: imageURL),
               let image = UIImage(data: data),
               let page = PDFPage(image: image),
               let documentURL = documentURL,
               let lastIndex = pdfView.document?.pageCount {
                DispatchQueue.main.async {
                    pdfView.document?.insert(page, at: lastIndex)
                    pdfView.document?.write(to: documentURL)
                    pdfView.go(to: page)
                    
                    LocalFileManager().updateThumbnail(forURL: documentURL)
                }
            }
        }
    }
  
  class func appendPage(images: [UIImage], pdfView: PDFView, documentURL: URL?) {
    guard let documentURL = documentURL else  { return }
    let currentIndex = pdfView.getCurrentPageIndex()
    
    for image in images {
      if let page = PDFPage(image: image) {
        pdfView.document?.insert(page, at: currentIndex)
      }
    }
    
    if let page = pdfView.visiblePages.first {
      pdfView.document?.write(to: documentURL)
      pdfView.go(to: page)
    }
    
    LocalFileManager().updateThumbnail(forURL: documentURL)
    
    pdfView.document = PDFDocument(url: documentURL)
    pdfView.layoutDocumentView()
  }
  
  class func deletePage(at index: Int, pdfDocument: PDFDocument?, documentURL: URL?, pdfView: PDFView!) {
    guard let pdfDocument = pdfDocument, let documentURL = documentURL, let pdfView = pdfView else  { return }
    pdfDocument.removePage(at: index)
    pdfDocument.write(to: documentURL)
    pdfView.document = pdfDocument
    pdfView.layoutDocumentView()
    LocalFileManager().updateThumbnail(forURL: documentURL)
    if pdfView.canGoToNextPage {
      pdfView.goToNextPage(pdfView)
    } else if pdfView.canGoToPreviousPage {
      pdfView.goToPreviousPage(pdfView)
    } else if pdfView.canGoToLastPage {
      pdfView.goToLastPage(pdfView)
    } else if pdfView.canGoForward {
      pdfView.goForward(pdfView)
    } else if pdfView.canGoBack {
      pdfView.goBack(pdfView)
    } else if pdfView.canGoToFirstPage {
      pdfView.goToFirstPage(pdfView)
    } else {
      pdfView.layoutDocumentView()
    }
  }
}
