import Foundation
import UIKit

private var _excludesFileExtensions = [String]()
private var excludesFilepaths: [URL]?

enum SortyFileType {
  case name, date
}

class LocalFileManager {
  
  let filesManager = FileManager.default
  
  var documentsURL: URL {
    return filesManager.urls(for: .documentDirectory, in: .userDomainMask)[0]
  }
  
  var scannerURL: URL {
      let iFinderURL = documentsURL.appendingPathComponent(AppConfigurator.Folders.AllDocument.rawValue, isDirectory: true)
    /// Check if file not exist and create
    if !filesManager.fileExists(atPath: iFinderURL.path) {
      try? filesManager.createDirectory(at: iFinderURL, withIntermediateDirectories: true, attributes: nil)
    }
    
    return iFinderURL
  }
    
    func getFolderUrl(folder: AppConfigurator.Folder) -> URL {
        let iFinderURL = documentsURL.appendingPathComponent(folder.savedName, isDirectory: true)
      /// Check if file not exist and create
      if !filesManager.fileExists(atPath: iFinderURL.path) {
        try? filesManager.createDirectory(at: iFinderURL, withIntermediateDirectories: true, attributes: nil)
      }
      
      return iFinderURL
    }
  
  /// Mapped for case insensitivity
  var excludesFileExtensions: [String]? {
    get {
      return _excludesFileExtensions.map { $0.lowercased() }
    }
    set {
      if let newValue = newValue {
        _excludesFileExtensions = newValue
      }
    }
  }
  
  func getItemsAtPath(_ path: URL) -> [URL]? {
    return try? filesManager.contentsOfDirectory(at: path, includingPropertiesForKeys: [], options: [])
  }
  
  func filesForDirectory(_ directoryPath: URL, sortType: SortyFileType) -> [File]  {
    var files = [File]()
    var fileURLs = [URL]()
    
    // Get contents
    do  {
      fileURLs = try self.filesManager.contentsOfDirectory(at: directoryPath, includingPropertiesForKeys: [], options: [.skipsHiddenFiles])
    } catch {
      return files
    }
    
    // Parse
    for fileURL in fileURLs {
      let file = File(fileURL: fileURL)
      
      if let excludesFileExtensions = excludesFileExtensions,
         let fileExtensions = file.fileExtension,
         excludesFileExtensions.contains(fileExtensions) {  continue }
      
      if let excludesFilepaths = excludesFilepaths, excludesFilepaths.contains(file.fileURL) {  continue }
      
      if !file.displayName.isEmpty && !file.displayName.lowercased().contains("(a document being saved") { files.append(file) }
      
      self.createThumbnailAndOriginalFolders(directoryPath)
      
      if file.fileURL.isPDF {
        let pdfThumbnailURL = getThumbnailFile(for: fileURL)
        
        if !self.filesManager.fileExists(atPath: pdfThumbnailURL.path) {
          switch deviceType {
          case .pad:
            if let pdfThumbnail = PDFManager.generatePdfThumbnail(of: CGSize(width: 225, height: 450), for: fileURL)?.toFullData {
              try? pdfThumbnail.write(to: pdfThumbnailURL)
            }
          default:
            if let pdfThumbnail = PDFManager.generatePdfThumbnail(of: CGSize(width: 120, height: 240), for: fileURL)?.toFullData {
              try? pdfThumbnail.write(to: pdfThumbnailURL)
            }
          }
        }
      }
    }
    
    switch sortType {
    case .date:
      return files.sorted { compare(date1: $0.fileURL.getFileAttributes?.fileCreationDate(),
                                    with: $1.fileURL.getFileAttributes?.fileCreationDate()) }
    case .name:
      return files.sorted { $0.displayName.lowercased() < $1.displayName.lowercased() }
    }
  }
  
  private func compare(date1: Date?, with date2: Date?) -> Bool {
    if let date1 = date1, let date2 = date2 {
      return date1 > date2
    }
    
    return false
  }
  
  func createThumbnailAndOriginalFolders(_ dirURL: URL) {
    self.create(folderAtPath: dirURL, withName: AppConfigurator.ThumbnailsFolderName)
  }
  
  func create(folderAtPath path: URL, withName name: String) {
    let newDirectory = path.appendingPathComponent(name)
    
    do {
      try filesManager.createDirectory(at: newDirectory, withIntermediateDirectories: true, attributes: nil)
      
    } catch {
      print("Cannot create folder at path, error = \(error.localizedDescription)")
    }
  }
  
  func removeThumbnail(for url: URL) throws {
    try filesManager.removeItem(at: getThumbnailFile(for: url))
  }
  
  func getThumbnailFile(for url: URL) -> URL {
    return url.deletingLastPathComponent().appendingPathComponent(AppConfigurator.ThumbnailsFolderName).appendingPathComponent(url.deletingPathExtension().lastPathComponent + ".jpg")
  }
  
  func getThumbnail(for pdfURL: URL) -> UIImage? {
    let imageURL = getThumbnailFile(for: pdfURL)
    return UIImage(contentsOfFile: imageURL.path)
  }
  
  func updateThumbnail(forURL fileURL: URL) {
    let pdfThumbnailURL = getThumbnailFile(for: fileURL)
    
    switch deviceType {
    case .pad:
      if let pdfThumbnail = PDFManager.generatePdfThumbnail(of: CGSize(width: 225, height: 450), for: fileURL)?.toFullData {
        try? pdfThumbnail.write(to: pdfThumbnailURL)
      }
    default:
      if let pdfThumbnail = PDFManager.generatePdfThumbnail(of: CGSize(width: 120, height: 240), for: fileURL)?.toFullData {
        try? pdfThumbnail.write(to: pdfThumbnailURL)
      }
    }
  }
  
  func renameFile(at url: URL, withName newName: String) throws {
    let fileURL = url.deletingLastPathComponent()
    let newFileName = "\(newName).pdf"
    try filesManager.moveItem(at: url, to: fileURL.appendingPathComponent(newFileName))
    try filesManager.removeItem(at: getThumbnailFile(for: url))
  }
}
